@@ -6,20 +6,21 @@ import sys
 from typing import Dict, List, Mapping, Optional, Pattern, Set, Tuple
 MYPY = False
 if MYPY:
-    from typing import ClassVar
+    from typing_extensions import Final
 
 from mypy import defaults
 from mypy.util import get_class_descriptors, replace_object_state
 
 
 class BuildType:
-    STANDARD = 0  # type: ClassVar[int]
-    MODULE = 1  # type: ClassVar[int]
-    PROGRAM_TEXT = 2  # type: ClassVar[int]
+    STANDARD = 0  # type: Final[int]
+    MODULE = 1  # type: Final[int]
+    PROGRAM_TEXT = 2  # type: Final[int]
 
 
 PER_MODULE_OPTIONS = {
     # Please keep this list sorted
+    "allow_untyped_globals",
     "always_false",
     "always_true",
     "check_untyped_defs",
@@ -39,19 +40,19 @@ PER_MODULE_OPTIONS = {
     "ignore_errors",
     "ignore_missing_imports",
     "local_partial_types",
+    "mypyc",
     "no_implicit_optional",
     "show_none_errors",
-    "strict_boolean",
     "strict_optional",
     "strict_optional_whitelist",
     "warn_no_return",
     "warn_return_any",
     "warn_unused_ignores",
-}
+}  # type: Final
 
 OPTIONS_AFFECTING_CACHE = ((PER_MODULE_OPTIONS |
                             {"quick_and_dirty", "platform", "bazel"})
-                           - {"debug_cache"})
+                           - {"debug_cache"})  # type: Final
 
 
 class Options:
@@ -78,7 +79,9 @@ class Options:
         self.follow_imports = 'normal'  # normal|silent|skip|error
         # Whether to respect the follow_imports setting even for stub files.
         # Intended to be used for disabling specific stubs.
-        self.follow_imports_for_stubs = False  # type: bool
+        self.follow_imports_for_stubs = False
+        # PEP 420 namespace packages
+        self.namespace_packages = False
 
         # disallow_any options
         self.disallow_any_generics = False
@@ -127,9 +130,6 @@ class Options:
         # Files in which to ignore all non-fatal errors
         self.ignore_errors = False
 
-        # Only allow booleans in conditions
-        self.strict_boolean = False
-
         # Apply strict None checking
         self.strict_optional = True
 
@@ -146,6 +146,9 @@ class Options:
         # Don't assume arguments with default values of None are Optional
         self.no_implicit_optional = False
 
+        # Suppress toplevel errors caused by missing annotations
+        self.allow_untyped_globals = False
+
         # Variable names considered True
         self.always_true = []  # type: List[str]
 
@@ -173,6 +176,10 @@ class Options:
         # Read cache files in fine-grained incremental mode (cache must include dependencies)
         self.use_fine_grained_cache = False
 
+        # Tune certain behaviors when being used as a front-end to mypyc. Set per-module
+        # in modules being compiled. Not in the config file or command line.
+        self.mypyc = False
+
         # Paths of user plugins
         self.plugins = []  # type: List[str]
 
@@ -211,6 +218,8 @@ class Options:
         # if they don't have __init__.py.
         self.package_root = []  # type: List[str]
         self.cache_map = {}  # type: Dict[str, Tuple[str, str]]
+        # Don't properly free objects on exit, just kill the current process.
+        self.fast_exit = False
 
     def snapshot(self) -> object:
         """Produce a comparable snapshot of this Option"""
@@ -328,7 +337,7 @@ class Options:
         parts = s.split('.')
         expr = re.escape(parts[0]) if parts[0] != '*' else '.*'
         for part in parts[1:]:
-            expr += re.escape('.' + part) if part != '*' else '(\..*)?'
+            expr += re.escape('.' + part) if part != '*' else r'(\..*)?'
         return re.compile(expr + '\\Z')
 
     def select_options_affecting_cache(self) -> Mapping[str, object]:
