@@ -11,7 +11,7 @@ An affected location is a string than can refer to a *target* (a non-nested
 function or method, or a module top level), a class, or a trigger (for
 recursively triggering other triggers).
 
-Here's an example represention of a simple dependency map (in format
+Here's an example representation of a simple dependency map (in format
 "<trigger> -> locations"):
 
   <m.A.g> -> m.f
@@ -90,27 +90,30 @@ from mypy.nodes import (
     Node, Expression, MypyFile, FuncDef, ClassDef, AssignmentStmt, NameExpr, MemberExpr, Import,
     ImportFrom, CallExpr, CastExpr, TypeVarExpr, TypeApplication, IndexExpr, UnaryExpr, OpExpr,
     ComparisonExpr, GeneratorExpr, DictionaryComprehension, StarExpr, PrintStmt, ForStmt, WithStmt,
-    TupleExpr, ListExpr, OperatorAssignmentStmt, DelStmt, YieldFromExpr, Decorator, Block,
+    TupleExpr, OperatorAssignmentStmt, DelStmt, YieldFromExpr, Decorator, Block,
     TypeInfo, FuncBase, OverloadedFuncDef, RefExpr, SuperExpr, Var, NamedTupleExpr, TypedDictExpr,
-    LDEF, MDEF, GDEF, FuncItem, TypeAliasExpr,
+    LDEF, MDEF, GDEF, TypeAliasExpr, NewTypeExpr, ImportAll, EnumCallExpr, AwaitExpr,
     op_methods, reverse_op_methods, ops_with_inplace_method, unary_op_methods
 )
 from mypy.traverser import TraverserVisitor
 from mypy.types import (
     Type, Instance, AnyType, NoneTyp, TypeVisitor, CallableType, DeletedType, PartialType,
     TupleType, TypeType, TypeVarType, TypedDictType, UnboundType, UninhabitedType, UnionType,
-    FunctionLike, ForwardRef, Overloaded
+    FunctionLike, ForwardRef, Overloaded, TypeOfAny
 )
-from mypy.server.trigger import make_trigger
+from mypy.server.trigger import make_trigger, make_wildcard_trigger
 from mypy.util import correct_relative_import
 from mypy.scope import Scope
+from mypy.typestate import TypeState
+from mypy.options import Options
 
 
 def get_dependencies(target: MypyFile,
                      type_map: Dict[Expression, Type],
-                     python_version: Tuple[int, int]) -> Dict[str, Set[str]]:
+                     python_version: Tuple[int, int],
+                     options: Options) -> Dict[str, Set[str]]:
     """Get all dependencies of a node, recursively."""
-    visitor = DependencyVisitor(type_map, python_version, target.alias_deps)
+    visitor = DependencyVisitor(type_map, python_version, target.alias_deps, options)
     target.accept(visitor)
     return visitor.map
 
@@ -147,7 +150,8 @@ class DependencyVisitor(TraverserVisitor):
     def __init__(self,
                  type_map: Dict[Expression, Type],
                  python_version: Tuple[int, int],
-                 alias_deps: 'DefaultDict[str, Set[str]]') -> None:
+                 alias_deps: 'DefaultDict[str, Set[str]]',
+                 options: Optional[Options] = None) -> None:
         self.scope = Scope()
         self.type_map = type_map
         self.python2 = python_version[0] == 2
@@ -164,14 +168,7 @@ class DependencyVisitor(TraverserVisitor):
         self.map = {}  # type: Dict[str, Set[str]]
         self.is_class = False
         self.is_package_init_file = False
-
-    # TODO (incomplete):
-    #   from m import *
-    #   await
-    #   protocols
-    #   metaclasses
-    #   functional enum
-    #   type variable with value restriction
+        self.options = options
 
     def visit_mypy_file(self, o: MypyFile) -> None:
         self.scope.enter_file(o.fullname())
@@ -188,18 +185,45 @@ class DependencyVisitor(TraverserVisitor):
                 signature = bind_self(o.type)  # type: Type
             else:
                 signature = o.type
-            for trigger in get_type_triggers(signature):
+            for trigger in self.get_type_triggers(signature):
                 self.add_dependency(trigger)
                 self.add_dependency(trigger, target=make_trigger(target))
         if o.info:
             for base in non_trivial_bases(o.info):
-                self.add_dependency(make_trigger(base.fullname() + '.' + o.name()))
+                # Base class __init__/__new__ doesn't generate a logical
+                # dependency since the override can be incompatible.
+                if not self.use_logical_deps() or o.name() not in ('__init__', '__new__'):
+                    self.add_dependency(make_trigger(base.fullname() + '.' + o.name()))
         self.add_type_alias_deps(self.scope.current_target())
         super().visit_func_def(o)
+        variants = set(o.expanded) - {o}
+        for ex in variants:
+            if isinstance(ex, FuncDef):
+                super().visit_func_def(ex)
         self.scope.leave()
 
     def visit_decorator(self, o: Decorator) -> None:
-        self.add_dependency(make_trigger(o.func.fullname()))
+        # We don't need to recheck outer scope for an overload, only overload itself.
+        # Also if any decorator is nested, it is not externally visible, so we don't need to
+        # generate dependency.
+        if not o.func.is_overload and self.scope.current_function_name() is None:
+            self.add_dependency(make_trigger(o.func.fullname()))
+        if self.use_logical_deps():
+            # Add logical dependencies from decorators to the function. For example,
+            # if we have
+            #     @dec
+            #     def func(): ...
+            # then if `dec` is unannotated, then it will "spoil" `func` and consequently
+            # all call sites, making them all `Any`.
+            for d in o.decorators:
+                tname = None  # type: Optional[str]
+                if isinstance(d, RefExpr) and d.fullname is not None:
+                    tname = d.fullname
+                if (isinstance(d, CallExpr) and isinstance(d.callee, RefExpr) and
+                        d.callee.fullname is not None):
+                    tname = d.callee.fullname
+                if tname is not None:
+                    self.add_dependency(make_trigger(tname), make_trigger(o.func.fullname()))
         super().visit_decorator(o)
 
     def visit_class_def(self, o: ClassDef) -> None:
@@ -211,20 +235,50 @@ class DependencyVisitor(TraverserVisitor):
         # Add dependencies to type variables of a generic class.
         for tv in o.type_vars:
             self.add_dependency(make_trigger(tv.fullname), target)
-        # Add dependencies to base types.
-        for base in o.info.bases:
-            self.add_type_dependencies(base, target=target)
-        if o.info.tuple_type:
-            self.add_type_dependencies(o.info.tuple_type, target=make_trigger(target))
-        if o.info.typeddict_type:
-            self.add_type_dependencies(o.info.typeddict_type, target=make_trigger(target))
-        # TODO: Add dependencies based on remaining TypeInfo attributes.
+        self.process_type_info(o.info)
         super().visit_class_def(o)
-        self.add_type_alias_deps(self.scope.current_target())
         self.is_class = old_is_class
-        info = o.info
+        self.scope.leave()
+
+    def visit_newtype_expr(self, o: NewTypeExpr) -> None:
+        if o.info:
+            self.scope.enter_class(o.info)
+            self.process_type_info(o.info)
+            self.scope.leave()
+
+    def process_type_info(self, info: TypeInfo) -> None:
+        target = self.scope.current_full_target()
+        for base in info.bases:
+            self.add_type_dependencies(base, target=target)
+        if info.tuple_type:
+            self.add_type_dependencies(info.tuple_type, target=make_trigger(target))
+        if info.typeddict_type:
+            self.add_type_dependencies(info.typeddict_type, target=make_trigger(target))
+        if info.declared_metaclass:
+            self.add_type_dependencies(info.declared_metaclass, target=make_trigger(target))
+        if info.is_protocol:
+            for base_info in info.mro[:-1]:
+                # We add dependencies from whole MRO to cover explicit subprotocols.
+                # For example:
+                #
+                #     class Super(Protocol):
+                #         x: int
+                #     class Sub(Super, Protocol):
+                #         y: int
+                #
+                # In this example we add <Super[wildcard]> -> <Sub>, to invalidate Sub if
+                # a new member is added to Super.
+                self.add_dependency(make_wildcard_trigger(base_info.fullname()),
+                                    target=make_trigger(target))
+                # More protocol dependencies are collected in TypeState._snapshot_protocol_deps
+                # after a full run or update is finished.
+
+        self.add_type_alias_deps(self.scope.current_target())
         for name, node in info.names.items():
             if isinstance(node.node, Var):
+                # Recheck Liskov if needed, self definitions are checked in the defining method
+                if node.node.is_initialized_in_class and has_user_bases(info):
+                    self.add_dependency(make_trigger(info.fullname() + '.' + name))
                 for base_info in non_trivial_bases(info):
                     # If the type of an attribute changes in a base class, we make references
                     # to the attribute in the subclass stale.
@@ -232,15 +286,47 @@ class DependencyVisitor(TraverserVisitor):
                                         target=make_trigger(info.fullname() + '.' + name))
         for base_info in non_trivial_bases(info):
             for name, node in base_info.names.items():
+                if self.use_logical_deps():
+                    # Skip logical dependency if an attribute is not overridden. For example,
+                    # in case of:
+                    #     class Base:
+                    #         x = 1
+                    #         y = 2
+                    #     class Sub(Base):
+                    #         x = 3
+                    # we skip <Base.y> -> <Child.y>, because even if `y` is unannotated it
+                    # doesn't affect precision of Liskov checking.
+                    if name not in info.names:
+                        continue
+                    # __init__ and __new__ can be overridden with different signatures, so no
+                    # logical depedency.
+                    if name in ('__init__', '__new__'):
+                        continue
                 self.add_dependency(make_trigger(base_info.fullname() + '.' + name),
                                     target=make_trigger(info.fullname() + '.' + name))
-            self.add_dependency(make_trigger(base_info.fullname() + '.__init__'),
-                                target=make_trigger(info.fullname() + '.__init__'))
-        self.scope.leave()
+            if not self.use_logical_deps():
+                # These dependencies are only useful for propagating changes --
+                # they aren't logical dependencies since __init__ and __new__ can be
+                # overridden with a different signature.
+                self.add_dependency(make_trigger(base_info.fullname() + '.__init__'),
+                                    target=make_trigger(info.fullname() + '.__init__'))
+                self.add_dependency(make_trigger(base_info.fullname() + '.__new__'),
+                                    target=make_trigger(info.fullname() + '.__new__'))
+                # If the set of abstract attributes change, this may invalidate class
+                # instantiation, or change the generated error message, since Python checks
+                # class abstract status when creating an instance.
+                #
+                # TODO: We should probably add this dependency only from the __init__ of the
+                #     current class, and independent of bases (to trigger changes in message
+                #     wording, as errors may enumerate all abstract attributes).
+                self.add_dependency(make_trigger(base_info.fullname() + '.(abstract)'),
+                                    target=make_trigger(info.fullname() + '.__init__'))
+                # If the base class abstract attributes change, subclass abstract
+                # attributes need to be recalculated.
+                self.add_dependency(make_trigger(base_info.fullname() + '.(abstract)'))
 
     def visit_import(self, o: Import) -> None:
         for id, as_id in o.ids:
-            # TODO: as_id
             self.add_dependency(make_trigger(id), self.scope.current_target())
 
     def visit_import_from(self, o: ImportFrom) -> None:
@@ -251,19 +337,29 @@ class DependencyVisitor(TraverserVisitor):
         for name, as_name in o.names:
             self.add_dependency(make_trigger(module_id + '.' + name))
 
+    def visit_import_all(self, o: ImportAll) -> None:
+        module_id, _ = correct_relative_import(self.scope.current_module_id(),
+                                               o.relative,
+                                               o.id,
+                                               self.is_package_init_file)
+        # The current target needs to be rechecked if anything "significant" changes in the
+        # target module namespace (as the imported definitions will need to be updated).
+        self.add_dependency(make_wildcard_trigger(module_id))
+
     def visit_block(self, o: Block) -> None:
         if not o.is_unreachable:
             super().visit_block(o)
 
     def visit_assignment_stmt(self, o: AssignmentStmt) -> None:
-        # TODO: Implement all assignment special forms, including these:
-        #   Enum
         rvalue = o.rvalue
         if isinstance(rvalue, CallExpr) and isinstance(rvalue.analyzed, TypeVarExpr):
-            # TODO: Support type variable value restriction
             analyzed = rvalue.analyzed
             self.add_type_dependencies(analyzed.upper_bound,
                                        target=make_trigger(analyzed.fullname()))
+            for val in analyzed.values:
+                self.add_type_dependencies(val, target=make_trigger(analyzed.fullname()))
+            # We need to re-analyze the definition if bound or value is deleted.
+            super().visit_call_expr(rvalue)
         elif isinstance(rvalue, CallExpr) and isinstance(rvalue.analyzed, NamedTupleExpr):
             # Depend on types of named tuple items.
             info = rvalue.analyzed.info
@@ -282,6 +378,11 @@ class DependencyVisitor(TraverserVisitor):
             assert info.typeddict_type is not None
             prefix = '%s.%s' % (self.scope.current_full_target(), info.name())
             self.add_type_dependencies(info.typeddict_type, target=make_trigger(prefix))
+        elif isinstance(rvalue, CallExpr) and isinstance(rvalue.analyzed, EnumCallExpr):
+            # Enum values are currently not checked, but for future we add the deps on them
+            for name, symnode in rvalue.analyzed.info.names.items():
+                if isinstance(symnode.node, Var) and symnode.node.type:
+                    self.add_type_dependencies(symnode.node.type)
         elif o.is_alias_def:
             assert len(o.lvalues) == 1
             lvalue = o.lvalues[0]
@@ -291,8 +392,11 @@ class DependencyVisitor(TraverserVisitor):
             if isinstance(typ, FunctionLike) and typ.is_type_obj():
                 class_name = typ.type_object().fullname()
                 self.add_dependency(make_trigger(class_name + '.__init__'))
+                self.add_dependency(make_trigger(class_name + '.__new__'))
             if isinstance(rvalue, IndexExpr) and isinstance(rvalue.analyzed, TypeAliasExpr):
                 self.add_type_dependencies(rvalue.analyzed.type)
+            elif typ:
+                self.add_type_dependencies(typ)
         else:
             # Normal assignment
             super().visit_assignment_stmt(o)
@@ -305,8 +409,30 @@ class DependencyVisitor(TraverserVisitor):
                 if isinstance(lvalue, TupleExpr):
                     self.add_attribute_dependency_for_expr(rvalue, '__iter__')
             if o.type:
-                for trigger in get_type_triggers(o.type):
+                for trigger in self.get_type_triggers(o.type):
                     self.add_dependency(trigger)
+        if self.use_logical_deps() and o.unanalyzed_type is None:
+            # Special case: for definitions without an explicit type like this:
+            #     x = func(...)
+            # we add a logical dependency <func> -> <x>, because if `func` is not annotated,
+            # then it will make all points of use of `x` unchecked.
+            if (isinstance(rvalue, CallExpr) and isinstance(rvalue.callee, RefExpr)
+                    and rvalue.callee.fullname is not None):
+                fname = None  # type: Optional[str]
+                if isinstance(rvalue.callee.node, TypeInfo):
+                    # use actual __init__ as a dependency source
+                    init = rvalue.callee.node.get('__init__')
+                    if init and isinstance(init.node, FuncBase):
+                        fname = init.node.fullname()
+                else:
+                    fname = rvalue.callee.fullname
+                if fname is None:
+                    return
+                for lv in o.lvalues:
+                    if isinstance(lv, RefExpr) and lv.fullname and lv.is_new_def:
+                        if lv.kind == LDEF:
+                            return  # local definitions don't generate logical deps
+                        self.add_dependency(make_trigger(fname), make_trigger(lv.fullname))
 
     def process_lvalue(self, lvalue: Expression) -> None:
         """Generate additional dependencies for an lvalue."""
@@ -317,12 +443,19 @@ class DependencyVisitor(TraverserVisitor):
                 # Assignment to an attribute in the class body, or direct assignment to a
                 # global variable.
                 lvalue_type = self.get_non_partial_lvalue_type(lvalue)
-                type_triggers = get_type_triggers(lvalue_type)
+                type_triggers = self.get_type_triggers(lvalue_type)
                 attr_trigger = make_trigger('%s.%s' % (self.scope.current_full_target(),
                                                        lvalue.name))
                 for type_trigger in type_triggers:
                     self.add_dependency(type_trigger, attr_trigger)
         elif isinstance(lvalue, MemberExpr):
+            if self.is_self_member_ref(lvalue) and lvalue.is_new_def:
+                node = lvalue.node
+                if isinstance(node, Var):
+                    info = node.info
+                    if info and has_user_bases(info):
+                        # Recheck Liskov for self definitions
+                        self.add_dependency(make_trigger(info.fullname() + '.' + lvalue.name))
             if lvalue.kind is None:
                 # Reference to a non-module attribute
                 if lvalue.expr not in self.type_map:
@@ -330,14 +463,22 @@ class DependencyVisitor(TraverserVisitor):
                     return
                 object_type = self.type_map[lvalue.expr]
                 lvalue_type = self.get_non_partial_lvalue_type(lvalue)
-                type_triggers = get_type_triggers(lvalue_type)
+                type_triggers = self.get_type_triggers(lvalue_type)
                 for attr_trigger in self.attribute_triggers(object_type, lvalue.name):
                     for type_trigger in type_triggers:
                         self.add_dependency(type_trigger, attr_trigger)
         elif isinstance(lvalue, TupleExpr):
             for item in lvalue.items:
                 self.process_lvalue(item)
-        # TODO: star lvalue
+        elif isinstance(lvalue, StarExpr):
+            self.process_lvalue(lvalue.expr)
+
+    def is_self_member_ref(self, memberexpr: MemberExpr) -> bool:
+        """Does memberexpr to refer to an attribute of self?"""
+        if not isinstance(memberexpr.expr, NameExpr):
+            return False
+        node = memberexpr.expr.node
+        return isinstance(node, Var) and node.is_self
 
     def get_non_partial_lvalue_type(self, lvalue: RefExpr) -> Type:
         if lvalue not in self.type_map:
@@ -368,10 +509,22 @@ class DependencyVisitor(TraverserVisitor):
 
     def visit_for_stmt(self, o: ForStmt) -> None:
         super().visit_for_stmt(o)
-        # __getitem__ is only used if __iter__ is missing but for simplicity we
-        # just always depend on both.
-        self.add_attribute_dependency_for_expr(o.expr, '__iter__')
-        self.add_attribute_dependency_for_expr(o.expr, '__getitem__')
+        if not o.is_async:
+            # __getitem__ is only used if __iter__ is missing but for simplicity we
+            # just always depend on both.
+            self.add_attribute_dependency_for_expr(o.expr, '__iter__')
+            self.add_attribute_dependency_for_expr(o.expr, '__getitem__')
+            if o.inferred_iterator_type:
+                if self.python2:
+                    method = 'next'
+                else:
+                    method = '__next__'
+                self.add_attribute_dependency(o.inferred_iterator_type, method)
+        else:
+            self.add_attribute_dependency_for_expr(o.expr, '__aiter__')
+            if o.inferred_iterator_type:
+                self.add_attribute_dependency(o.inferred_iterator_type, '__anext__')
+
         self.process_lvalue(o.index)
         if isinstance(o.index, TupleExpr):
             # Process multiple assignment to index variables.
@@ -386,8 +539,12 @@ class DependencyVisitor(TraverserVisitor):
     def visit_with_stmt(self, o: WithStmt) -> None:
         super().visit_with_stmt(o)
         for e in o.expr:
-            self.add_attribute_dependency_for_expr(e, '__enter__')
-            self.add_attribute_dependency_for_expr(e, '__exit__')
+            if not o.is_async:
+                self.add_attribute_dependency_for_expr(e, '__enter__')
+                self.add_attribute_dependency_for_expr(e, '__exit__')
+            else:
+                self.add_attribute_dependency_for_expr(e, '__aenter__')
+                self.add_attribute_dependency_for_expr(e, '__aexit__')
         if o.target_type:
             self.add_type_dependencies(o.target_type)
 
@@ -409,16 +566,17 @@ class DependencyVisitor(TraverserVisitor):
 
         # If this is a reference to a type, generate a dependency to its
         # constructor.
-        # TODO: avoid generating spurious dependencies for isinstancce checks,
+        # TODO: avoid generating spurious dependencies for isinstance checks,
         # except statements, class attribute reference, etc, if perf problem.
         typ = self.type_map.get(o)
         if isinstance(typ, FunctionLike) and typ.is_type_obj():
             class_name = typ.type_object().fullname()
             self.add_dependency(make_trigger(class_name + '.__init__'))
+            self.add_dependency(make_trigger(class_name + '.__new__'))
 
     def visit_name_expr(self, o: NameExpr) -> None:
         if o.kind == LDEF:
-            # We don't track depdendencies to local variables, since they
+            # We don't track dependencies to local variables, since they
             # aren't externally visible.
             return
         if o.kind == MDEF:
@@ -440,6 +598,43 @@ class DependencyVisitor(TraverserVisitor):
                 return
             typ = self.type_map[e.expr]
             self.add_attribute_dependency(typ, e.name)
+            if self.use_logical_deps() and isinstance(typ, AnyType):
+                name = self.get_unimported_fullname(e, typ)
+                if name is not None:
+                    # Generate a logical dependency from an unimported
+                    # definition (which comes from a missing module).
+                    # Example:
+                    #     import missing  # "missing" not in build
+                    #
+                    #     def g() -> None:
+                    #         missing.f()  # Generate dependency from "missing.f"
+                    self.add_dependency(make_trigger(name))
+
+    def get_unimported_fullname(self, e: MemberExpr, typ: AnyType) -> Optional[str]:
+        """If e refers to an unimported definition, infer the fullname of this.
+
+        Return None if e doesn't refer to an unimported definition or if we can't
+        determine the name.
+        """
+        suffix = ''
+        # Unwrap nested member expression to handle cases like "a.b.c.d" where
+        # "a.b" is a known reference to an unimported module. Find the base
+        # reference to an unimported module (such as "a.b") and the name suffix
+        # (such as "c.d") needed to build a full name.
+        while typ.type_of_any == TypeOfAny.from_another_any and isinstance(e.expr, MemberExpr):
+            suffix = '.' + e.name + suffix
+            e = e.expr
+            if e.expr not in self.type_map:
+                return None
+            obj_type = self.type_map[e.expr]
+            if not isinstance(obj_type, AnyType):
+                # Can't find the base reference to the unimported module.
+                return None
+            typ = obj_type
+        if typ.type_of_any == TypeOfAny.from_unimported_type and typ.missing_import_name:
+            # Infer the full name of the unimported definition.
+            return typ.missing_import_name + '.' + e.name + suffix
+        return None
 
     def visit_super_expr(self, e: SuperExpr) -> None:
         super().visit_super_expr(e)
@@ -502,8 +697,6 @@ class DependencyVisitor(TraverserVisitor):
     def add_operator_method_dependency_for_type(self, typ: Type, method: str) -> None:
         # Note that operator methods can't be (non-metaclass) methods of type objects
         # (that is, TypeType objects or Callables representing a type).
-        # TODO: TypedDict
-        # TODO: metaclasses
         if isinstance(typ, TypeVarType):
             typ = typ.upper_bound
         if isinstance(typ, TupleType):
@@ -514,6 +707,11 @@ class DependencyVisitor(TraverserVisitor):
         elif isinstance(typ, UnionType):
             for item in typ.items:
                 self.add_operator_method_dependency_for_type(item, method)
+        elif isinstance(typ, FunctionLike) and typ.is_type_obj():
+            self.add_operator_method_dependency_for_type(typ.fallback, method)
+        elif isinstance(typ, TypeType):
+            if isinstance(typ.item, Instance) and typ.item.type.metaclass_type is not None:
+                self.add_operator_method_dependency_for_type(typ.item.type.metaclass_type, method)
 
     def visit_generator_expr(self, e: GeneratorExpr) -> None:
         super().visit_generator_expr(e)
@@ -533,6 +731,10 @@ class DependencyVisitor(TraverserVisitor):
         super().visit_yield_from_expr(e)
         self.add_iter_dependency(e.expr)
 
+    def visit_await_expr(self, e: AwaitExpr) -> None:
+        super().visit_await_expr(e)
+        self.add_attribute_dependency_for_expr(e.expr, '__await__')
+
     # Helpers
 
     def add_type_alias_deps(self, target: str) -> None:
@@ -566,7 +768,7 @@ class DependencyVisitor(TraverserVisitor):
         """
         # TODO: Use this method in more places where get_type_triggers() + add_dependency()
         #       are called together.
-        for trigger in get_type_triggers(typ):
+        for trigger in self.get_type_triggers(typ):
             self.add_dependency(trigger, target)
 
     def add_attribute_dependency(self, typ: Type, name: str) -> None:
@@ -586,15 +788,21 @@ class DependencyVisitor(TraverserVisitor):
             return [make_trigger(member)]
         elif isinstance(typ, FunctionLike) and typ.is_type_obj():
             member = '%s.%s' % (typ.type_object().fullname(), name)
-            return [make_trigger(member)]
+            triggers = [make_trigger(member)]
+            triggers.extend(self.attribute_triggers(typ.fallback, name))
+            return triggers
         elif isinstance(typ, UnionType):
             targets = []
             for item in typ.items:
                 targets.extend(self.attribute_triggers(item, name))
             return targets
         elif isinstance(typ, TypeType):
-            # TODO: Metaclass attribute lookup
-            return self.attribute_triggers(typ.item, name)
+            triggers = self.attribute_triggers(typ.item, name)
+            if isinstance(typ.item, Instance) and typ.item.type.metaclass_type is not None:
+                triggers.append(make_trigger('%s.%s' %
+                                             (typ.item.type.metaclass_type.type.fullname(),
+                                              name)))
+            return triggers
         else:
             return []
 
@@ -608,42 +816,54 @@ class DependencyVisitor(TraverserVisitor):
         if typ:
             self.add_attribute_dependency(typ, '__iter__')
 
+    def use_logical_deps(self) -> bool:
+        return self.options is not None and self.options.logical_deps
 
-def get_type_triggers(typ: Type) -> List[str]:
+    def get_type_triggers(self, typ: Type) -> List[str]:
+        return get_type_triggers(typ, self.use_logical_deps())
+
+
+def get_type_triggers(typ: Type, use_logical_deps: bool) -> List[str]:
     """Return all triggers that correspond to a type becoming stale."""
-    return typ.accept(TypeTriggersVisitor())
+    return typ.accept(TypeTriggersVisitor(use_logical_deps))
 
 
 class TypeTriggersVisitor(TypeVisitor[List[str]]):
-    def __init__(self) -> None:
+    def __init__(self, use_logical_deps: bool) -> None:
         self.deps = []  # type: List[str]
+        self.use_logical_deps = use_logical_deps
+
+    def get_type_triggers(self, typ: Type) -> List[str]:
+        return get_type_triggers(typ, self.use_logical_deps)
 
     def visit_instance(self, typ: Instance) -> List[str]:
         trigger = make_trigger(typ.type.fullname())
         triggers = [trigger]
         for arg in typ.args:
-            triggers.extend(get_type_triggers(arg))
+            triggers.extend(self.get_type_triggers(arg))
         return triggers
 
     def visit_any(self, typ: AnyType) -> List[str]:
+        if typ.missing_import_name is not None:
+            return [make_trigger(typ.missing_import_name)]
         return []
 
     def visit_none_type(self, typ: NoneTyp) -> List[str]:
         return []
 
     def visit_callable_type(self, typ: CallableType) -> List[str]:
-        # TODO: generic callables
-        # TODO: fallback?
         triggers = []
         for arg in typ.arg_types:
-            triggers.extend(get_type_triggers(arg))
-        triggers.extend(get_type_triggers(typ.ret_type))
+            triggers.extend(self.get_type_triggers(arg))
+        triggers.extend(self.get_type_triggers(typ.ret_type))
+        # fallback is a metaclass type for class objects, and is
+        # processed separately.
         return triggers
 
     def visit_overloaded(self, typ: Overloaded) -> List[str]:
         triggers = []
         for item in typ.items():
-            triggers.extend(get_type_triggers(item))
+            triggers.extend(self.get_type_triggers(item))
         return triggers
 
     def visit_deleted_type(self, typ: DeletedType) -> List[str]:
@@ -655,28 +875,37 @@ class TypeTriggersVisitor(TypeVisitor[List[str]]):
     def visit_tuple_type(self, typ: TupleType) -> List[str]:
         triggers = []
         for item in typ.items:
-            triggers.extend(get_type_triggers(item))
-        triggers.extend(get_type_triggers(typ.fallback))
+            triggers.extend(self.get_type_triggers(item))
+        triggers.extend(self.get_type_triggers(typ.fallback))
         return triggers
 
     def visit_type_type(self, typ: TypeType) -> List[str]:
-        return get_type_triggers(typ.item)
+        triggers = self.get_type_triggers(typ.item)
+        if not self.use_logical_deps:
+            old_triggers = triggers[:]
+            for trigger in old_triggers:
+                triggers.append(trigger.rstrip('>') + '.__init__>')
+                triggers.append(trigger.rstrip('>') + '.__new__>')
+        return triggers
 
     def visit_forwardref_type(self, typ: ForwardRef) -> List[str]:
         assert False, 'Internal error: Leaked forward reference object {}'.format(typ)
 
     def visit_type_var(self, typ: TypeVarType) -> List[str]:
-        # TODO: bound (values?)
         triggers = []
         if typ.fullname:
             triggers.append(make_trigger(typ.fullname))
+        if typ.upper_bound:
+            triggers.extend(self.get_type_triggers(typ.upper_bound))
+        for val in typ.values:
+            triggers.extend(self.get_type_triggers(val))
         return triggers
 
     def visit_typeddict_type(self, typ: TypedDictType) -> List[str]:
         triggers = []
         for item in typ.items.values():
-            triggers.extend(get_type_triggers(item))
-        triggers.extend(get_type_triggers(typ.fallback))
+            triggers.extend(self.get_type_triggers(item))
+        triggers.extend(self.get_type_triggers(typ.fallback))
         return triggers
 
     def visit_unbound_type(self, typ: UnboundType) -> List[str]:
@@ -688,7 +917,7 @@ class TypeTriggersVisitor(TypeVisitor[List[str]]):
     def visit_union_type(self, typ: UnionType) -> List[str]:
         triggers = []
         for item in typ.items:
-            triggers.extend(get_type_triggers(item))
+            triggers.extend(self.get_type_triggers(item))
         return triggers
 
 
@@ -697,9 +926,15 @@ def non_trivial_bases(info: TypeInfo) -> List[TypeInfo]:
             if base.fullname() != 'builtins.object']
 
 
+def has_user_bases(info: TypeInfo) -> bool:
+    # TODO: skip everything from typeshed?
+    return any(base.module_name not in ('builtins', 'typing', 'enum') for base in info.mro[1:])
+
+
 def dump_all_dependencies(modules: Dict[str, MypyFile],
                           type_map: Dict[Expression, Type],
-                          python_version: Tuple[int, int]) -> None:
+                          python_version: Tuple[int, int],
+                          options: Options) -> None:
     """Generate dependencies for all interesting modules and print them to stdout."""
     all_deps = {}  # type: Dict[str, Set[str]]
     for id, node in modules.items():
@@ -708,9 +943,10 @@ def dump_all_dependencies(modules: Dict[str, MypyFile],
         if id in ('builtins', 'typing') or '/typeshed/' in node.path:
             continue
         assert id == node.fullname()
-        deps = get_dependencies(node, type_map, python_version)
+        deps = get_dependencies(node, type_map, python_version, options)
         for trigger, targets in deps.items():
             all_deps.setdefault(trigger, set()).update(targets)
+    TypeState.add_all_protocol_deps(all_deps)
 
     for trigger, targets in sorted(all_deps.items(), key=lambda x: x[0]):
         print(trigger)
